# Generated by ariadne-codegen
# Source: ../../../shared/graphql/schema.graphql

from datetime import datetime
from typing import Annotated, Any, List, Optional, Union

from pydantic import Field, PlainSerializer

from .base_model import BaseModel, Upload


class UserInput(BaseModel):
    name: str
    email: str
    role: str
    extra_roles: Optional[List[str]] = Field(alias="extraRoles", default=None)
